//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gertespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "AVRChip.h"
#include "Conf.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum
{   DO_CONTINUE,
    DO_RESTART,
    DO_ENTER_APPL
} DoNext;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16           prgLdrMode __attribute__ ((section (".noinit")));
static volatile UInt16  msecTicks                                  =   0;
static volatile UInt08  busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY+1]    = { 0 };
static volatile UInt08  busRxQueueHead                             =   0;
static volatile UInt08  busRxQueueTail                             =   0;
static volatile UInt08  serRxQueue[CFG_SERIAL_RX_QUEUE_CAPACITY+1] = { 0 };
static volatile UInt08  serRxQueueHead                             =   0;
static volatile UInt08  serRxQueueTail                             =   0;
static          DoNext  doNext                                     = DO_CONTINUE;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
Restart(void)
{
    DisableInterrupts();
    prgLdrMode = CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
EnterAppl(void)
{
    DisableInterrupts();
    prgLdrMode = !CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//                          (X)                                                         //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                    /\  ja                                              //
//                   | Power-On-Reset |--------------------+                            //
//                    \______________/                     |                            //
//                           | nein                        |                            //
//                           |                             |                            //
//                          \|/                           \|/                           //
//                    /\ nein         /\ ja                 //
//                   | External-Reset |----------->| Watchdog-Reset |------+            //
//                    \______________/              \______________/       |            //
//                           | ja                          | nein          |            //
//                           |                             |               |            //
//                          \|/                           \|/              |            //
//                +----------------------+     +-----------------------+   |            //
//                | mode = PGMLDR_MARKER |     | mode = !PGMLDR_MARKER |   |            //
//                +----------------------+     +-----------------------+   |            //
//                           |                             |               |            //
//                           |<----------------------------+---------------+            //
//                           |                                                          //
//                          \|/                                                         //
//                /\  nein                                         //
//               | mode != PGMLDR_MARKER |-----------------+                            //
//                \_____________________/                  |                            //
//                           | ja                          |                            //
//                           |                             |                            //
//                          \|/                            |                            //
//                  /\  nein              |                            //
//                 | Read(0) != 0xFFFF |------------------>|                            //
//                  \_________________/                    |                            //
//                           |  ja                         |                            //
//                           |                             |                            //
//                   ________|________            _________|__________                  //
//                   Enter Application            Enter Program Loader                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked)) __attribute__((section(".init3"))) void
PreStart(void)
{
    UInt08 mcusr = MCUSR;
    MCUSR = 0;
    UInt16 mode = prgLdrMode;
    wdt_enable(WDTO_2S);
    if (mcusr & BV(PORF))
    {   mode = !CFG_PGMLDR_MARKER;
    }
    else
    {   if (mcusr & BV(EXTRF))
        {   mode = CFG_PGMLDR_MARKER;
        }
        else
        {   if (!(mcusr & BV(WDRF)))
            {   mode = !CFG_PGMLDR_MARKER;
            }
        }
    }
    if (mode != CFG_PGMLDR_MARKER && pgm_read_word_near(0) != 0xFFFF) GOTO(0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
    UInt08 byte = UDR0;
    UInt08 tail = busRxQueueTail+1;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead) return;
    busRxQueue[busRxQueueTail] = byte;
    busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART1_RX_vect)
{
    UInt08 status = UCSR1A;
    UInt08 byte   = UDR1;
    if (status & (BV(FE1)|BV(DOR1))) return;
    UInt08 tail = serRxQueueTail+1;
    if (tail == CARDINALITY(serRxQueue)) tail = 0;
    if (tail == serRxQueueHead) return;
    serRxQueue[serRxQueueTail] = byte;
    serRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(__vector_default) __attribute__((naked));
ISR(__vector_default)
{
    // Programmlader Einstiegspunkt
    Restart();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Restart(void)
{
    doNext = DO_RESTART;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterAppl(void)
{
    doNext = DO_ENTER_APPL;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceiveByte(UInt08* byte)
{
    if (busRxQueueHead == busRxQueueTail)
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = busRxQueue[busRxQueueHead];
        busRxQueueHead++;
        if (busRxQueueHead == CARDINALITY(busRxQueue)) busRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR1A, UDRE1);
    UDR1 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_SerReceiveByte(UInt08* byte)
{
    if (serRxQueueHead == serRxQueueTail)
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = serRxQueue[serRxQueueHead];
        serRxQueueHead++;
        if (serRxQueueHead == CARDINALITY(serRxQueue)) serRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(UInt16 adr, void* mem, UInt16 size)
{
    eeprom_read_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(UInt16 adr, const void* mem, UInt16 size)
{
    eeprom_update_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _L_,    // PA.0: ---        :
                _L_,    // PA.1: ---        :
                _L_,    // PA.2: ---        :
                _L_,    // PA.3: ---        :
                _L_,    // PA.4: ---        :
                _T_,    // PA.5: GPIO2      : Auf H3/Pin5 bereitgestellter Mehrzweck-Anschluss 2
                _L_,    // PA.6: ---        :
                _T_);   // PA.7: MON        : Monitor-Taste, L = gedrckt
    ConfPort(B, _L_,    // PB.0: ---        :
                _L_,    // PB.1: ---        :
                _L_,    // PB.2: ---        :
                _L_,    // PB.3: SER_ACT_LED: LED zur Anzeige serieller Aktivitt, H = Ein
                _L_,    // PB.4: LED_INFO   : Info-LED, H = Ein
                _P_,    // PB.5: ---        : [ISP]
                _P_,    // PB.6: ---        : [ISP]
                _P_);   // PB.7: ---        : [ISP]
    ConfPort(C, _L_,    // PC.0: ---        :
                _L_,    // PC.1: ---        :
                _L_,    // PC.2: ---        : [JTAG]
                _L_,    // PC.3: ---        : [JTAG]
                _L_,    // PC.4: ---        : [JTAG]
                _L_,    // PC.5: ---        : [JTAG]
                _L_,    // PC.6: ---        :
                _L_);   // PC.7: ---        :
    ConfPort(D, _P_,    // PD.0: RXD_BUS    : RXD von Bus-Treiber zu Controller
                _P_,    // PD.1: TXD_BUS    : TXD von Controller  zu Bus-Treiber
                _P_,    // PD.2: TXD_USB    : TXD von USB-Chip zu  Controller
                _P_,    // PD.3: RXD_USB    : RXD von Controller zu USB-Chip
                _L_,    // PD.4: ---        :
                _T_,    // PD.5: GPIO1      : Auf H3/Pin5 bereitgestellter Mehrzweck-Anschluss 1
                _L_,    // PD.6: ---        :
                _H_);   // PD.7: PWRON      : Auf H3/Pin1 durchgereichte USB-5V Spannung, L = Ein
    // Interrupt-Vektortabelle in den Bootloader-Bereich verlegen
    asm volatile(    "push  R22      \n\t"
                     "IN    R22, %0  \n\t"
                     "SBR   R22, %1  \n\t"
                     "OUT   %0, R22; \n\t"
                     "CBR   R22, %1  \n\t"
                     "SBR   R22, %2  \n\t"
                     "OUT   %0, r22; \n\t"
                     "POP   R22      \n\t"
                    :
                    :"M" (_SFR_IO_ADDR(MCUCR)), "i" (BV(IVCE)), "i" (BV(IVSEL))
                    :"r22");
    Freeze100us();
    // 16-Bit Timer 1 ohne IRQ zhlen lassen (F_CPU1024) --------------------------------
    TCCR1B = BV(CS12) | BV(CS10);
    // USART 0 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(RXEN0) | BV(TXEN0);
    UBRR0  = F_CPU/16L/19200-1;
    // USART 1 57600 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR1A = 0;
    UCSR1B = BV(RXCIE1) | BV(RXEN1) | BV(TXEN1);
    UBRR1  = F_CPU/16L/57600-1;
    // Systemzustandsleuchte konfigurieren
    ConfPin(CFG_SYSTEM_LED, _L_);
    // -----------------------------------------------------------------------------------
    EnableInterrupts();
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_KeepRunning(void)
{
    return doNext == DO_CONTINUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    wdt_reset();
    UInt16 ticks = TIMERTICKS/4;
    #if CFG_LED_LOW_ACTIVE == true
        ChgPin(CFG_SYSTEM_LED,  (ticks & 0x0640));
    #else
        ChgPin(CFG_SYSTEM_LED, !(ticks & 0x0640));
    #endif
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    switch (doNext)
    {   case DO_RESTART:    Restart();
        case DO_ENTER_APPL: EnterAppl();
        default:            doNext = DO_RESTART;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
