//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Cy4NET Programmlader Typ 1                                                          //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include "Conf.h"
#include "TypeDefs.h"
#include "AVRChip.h"
#include "Board.h"
#include "Cy4Message.h"
#include "Cy4TypePlate.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen & Definitionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#if   SPM_PAGESIZE == 32
    #define CFG_PAGE_SIZE 5
#elif SPM_PAGESIZE == 64
    #define CFG_PAGE_SIZE 6
#elif SPM_PAGESIZE == 128
    #define CFG_PAGE_SIZE 7
#elif SPM_PAGESIZE == 256
    #define CFG_PAGE_SIZE 8
#else
    #error Unsupported SPM_PAGESIZE
#endif

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen & Definitionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define PGMLDR_START   (FLASHEND-(CFG_PGMLDR_SIZE*1024)+1)  // Programmlader Startadresse
#define PAYLOAD_CAPACITY       64 // Kapazitt der PDU DAT Sektion: Fest 64 [Bytes]
#define PDU_DPS              0xAA   // Datagramm-Preamble-Sequence:   0b10101010
#define PDU_DIB              0x30   // Datagramm-Identification-Bits: 0b110000
#define PDU_DTF_P            BV(0)  // Datagram type flag: P (Payload)
#define PDU_DTF_D            BV(1)  // Datagram type flag: D (Direction)

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16      msecTicks              = 0;
static          UInt08      zero[1<<CFG_PAGE_SIZE] = { 0 };
static          UInt08      page[1<<CFG_PAGE_SIZE] = { 0 };
static          bool        progInProg             = false;
static          UInt16      ramPageNo              = 0xFFFF;
static          Cy4Address  selfAdr                = CY4ADDRESS_INVALID;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SaveSelfAddress(Cy4Address address)
{
    selfAdr = address;
    UInt08  data[4];
    data[0] =  address.net;
    data[1] =  address.dev;
    data[2] = ~address.net;
    data[3] = ~address.dev;
    BOARD_EEPSave(CFG_EEP_CY4_ADDRESS_PARTITION, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
LoadSelfAddress(void)
{
    UInt08  data[4];

    BOARD_EEPLoad(CFG_EEP_CY4_ADDRESS_PARTITION, data, sizeof(data));
    UInt08 invData[2] = { ~data[2], ~data[3] };
    if (data[0] == invData[0] && data[1] == invData[1])
    {   selfAdr.net = data[0];
        selfAdr.dev = data[1];
    }
    else
    {   selfAdr.net = CFG_DEFAULT_CY4_ADDRESS_NET;
        selfAdr.dev = CFG_DEFAULT_CY4_ADDRESS_DEV;
        SaveSelfAddress(selfAdr);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ReadFlashPage(UInt16 pageNo)
{
    UInt08*  buffer;

    if (pageNo == 0) buffer = zero;
    else             buffer = page;
    UInt16 pageAdr = (UInt16)pageNo << CFG_PAGE_SIZE;
    for (UInt16 i=0; i<(1<<CFG_PAGE_SIZE); i++) buffer[i] = pgm_read_byte(pageAdr+i);
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
WriteFlashPage(UInt16 pageNo)
{
    UInt08*  buffer;

    if (pageNo == 0) buffer = zero;
    else             buffer = page;
    UInt16 pageAdr = (UInt16)pageNo << CFG_PAGE_SIZE;
    // Erase
    Atomic
    {   boot_page_erase(pageAdr);
        boot_spm_busy_wait();
        // Write
        for (UInt16 i=0; i<(1<<CFG_PAGE_SIZE); i+=2)
        {   UInt16 w = ((UInt16)(buffer[i+1])<< 8) | buffer[i];
            boot_page_fill(pageAdr+i, w);
        }
        boot_page_write(pageAdr);
        boot_spm_busy_wait();
        boot_rww_enable();
    }
    // Verify
    for (UInt16 i=0; i<(1<<CFG_PAGE_SIZE); i++)
    {   if (buffer[i] != pgm_read_byte(pageAdr+i))
        {   return false;
        }
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
UpdateCRC(UInt08 data, UInt08* crc)
{
    *crc ^= data;
    for (UInt08 i=0; i<8; i++) *crc = (*crc >> 1) ^ ((*crc & 1) ? 0x8C : 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintUInt08Hex(char** text, UInt08 value)
{
    for (UInt08 i=0; i<2; i++)
    {   UInt08 n = (unsigned char)((value>>4) & 0xF);
        if (n < 10) **text =(char)('0'    + n);
        else        **text =(char)('A'-10 + n);
        (*text)++;
        value<<=4;
    }
    **text = '\0';
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
ParseUInt08Hex(const char** text, UInt08* value)
{
    const char* str = *text;
    *value = 0;
    for (UInt08 i=0; i<2; i++)
    {   UInt08 c = (UInt08)(*str);
        if (c >= 'a') c -= 'a' - 'A';  // Grobuchstaben
        UInt08 n;
        if      (c >= '0' && c <= '9') n =  0 + c - '0';
        else if (c >= 'A' && c <= 'F') n = 10 + c - 'A';
        else                           return false;
        *value = *value*16 + n;
        str++;
    }
    *text = str;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
HasPayload(const UInt08* pdu)
{
    return !!(pdu[1] & PDU_DTF_P);
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
BusGetRequest(Cy4Message* message)
{
    static UInt08  pdu[10+PAYLOAD_CAPACITY+1] = { 0 };
    static UInt08  index                      =   0;
    static UInt08  count                      =  10;
    static UInt08  crc                        =   0;
    static UInt16  timeout                    =   0;

    message->valid = false;
    pdu[0] = !PDU_DPS;
    // (1) Datagramm empfangen
    UInt08  byte;
    while (BOARD_BusReceiveByte(&byte))
    {   timeout = BOARD_SetTimeout(CFG_BUS_INTERBYTE_TIMEOUT+1);
        switch (index)
        {   case 0: // Prfen, ob Datagramm mit DPS beginnt
                if (byte != PDU_DPS) continue;
                count = 10;
                crc = 0xFF;
                break;
            case 1: // Prfen, ob DTF passt
                if ((byte>>2) != PDU_DIB)
                {   index = 0;
                    continue;
                }
                break;
            case 9: // Nutzlast-Grenze prfen und Datagrammlnge bestimmen
                if (byte > PAYLOAD_CAPACITY)
                {   index = 0;
                    continue;
                }
                if (HasPayload(pdu)) count += byte;
                break;
            default:
                ;
        }
        if (index > 0) pdu[index] = byte;
        UpdateCRC(byte, &crc);
        if (index == count)
        {   index = 0;
            if (crc == 0) pdu[0] = PDU_DPS;
            break;
        }
        index++;
    }
    if (BOARD_Timeout(timeout)) index = 0;
    if (pdu[0] != PDU_DPS) return false;
    // (2) Anfrage-Nachricht aus Datagramm erzeugen
    message->valid = pdu[2] == selfAdr.net && pdu[3] == selfAdr.dev;
    if (pdu[1] == 0xC1 || pdu[1] == 0xC2) message->access = CY4MESSAGE_READ;
    else                                  message->access = CY4MESSAGE_WRITE;
    message->address.net = pdu[4];
    message->address.dev = pdu[5];
    message->location    = ((UInt32)pdu[6]<<16) + ((UInt16)pdu[7]<<8) + pdu[8];
    message->count       = pdu[9];
    if (HasPayload(pdu))
    {    for (UInt08 i=0; i<message->count; i++) message->data[i] = pdu[10+i];
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
BusPutResponse(const Cy4Message* message)
{
    if (!message->valid) return;
    // (1) Datagramm aus Antwort-Nachricht erzeugen
    UInt08 dtf;
    if (message->access == CY4MESSAGE_WRITE) dtf = 0xC0;   // DTF: Write-Response
    else                                     dtf = 0xC1;   // DTP: Read-Response
    UInt08  pdu[10+PAYLOAD_CAPACITY+1];
    pdu[0] = PDU_DPS;
    pdu[1] = dtf;
    pdu[2] = message->address.net;
    pdu[3] = message->address.dev;
    pdu[4] = selfAdr.net;
    pdu[5] = selfAdr.dev;
    pdu[6] = (UInt08)(message->location >> 16);
    pdu[7] = (UInt08)(message->location >>  8);
    pdu[8] = (UInt08)(message->location >>  0);
    pdu[9] = message->count;
    UInt08 crc = 0xFF;
    UInt08 i;
    for (i=0; i<10; i++) UpdateCRC(pdu[i], &crc);
    if (HasPayload(pdu))
    {   for (i=0; i<pdu[9]; i++)
        {   pdu[10+i] = message->data[i];
            UpdateCRC(pdu[10+i], &crc);
        }
        i += 10;
    }
    pdu[i] = crc;
    // (2) Datagramm senden
    UInt08  n;
    if (HasPayload(pdu)) n = 11 + pdu[9];
    else                 n = 11;
    for (i=0; i<n; i++) BOARD_BusTransmitByte(pdu[i]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//   Cy4 ASCII PDU Synopsis:                                                            //
//   -----------------------                                                            //
//                                      XX XX XX XXXXXX XX XX XX ... XX XX              //
//                                      |  |  |    |    |  |  |      |  |               //
//                                      |  |  |    |    |  |  |      |  |               //
//           DTF:   "C0"   ..   "C5" ---+  |  |    |    |  |  |      |  |               //
//           NET:   "00"   ..   "FF" ------+  |    |    |  |  |      |  |               //
//           DEV:   "00"   ..   "FF" ---------+    |    |  |  |      |  |               //
//           LOC: "000000" .. "FFFFFF" ------------+    |  |  |      |  |               //
//           CNT:   "00"   ..   "40" -------------------+  |  |      |  |               //
//           DAT 1: "00"   ..   "FF" ----------------------+  |      |  |               //
//           DAT 2: "00"   ..   "FF" -------------------------+      |  |               //
//            :      ::    ..    ::                                  |  |               //
//           DAT n: "00"   ..   "FF" --------------------------------+  |               //
//           CRC:   "00"   ..   "FF" -----------------------------------+               //
//                                                                                      //
//                                                                                      //
//    DTF: Datagramm Type Flags Byte:                                                  //
//                                                                                      //
//           7 6 5 4 3 2 1 0                                                            //
//          +-+-+-+-+-+-+-+-+                                                           //
//          |1|1|0|0|0|F|D|P|                                                           //
//          +-+-+-+-+-+-+-+-+                                                           //
//                     | | |                                                            //
//                     | | +--- P: Payload flag:   1 = Has payload  , 0 = No payload    //
//                     | +----- D: Direction flag: 1 = Request      , 0 = Response      //
//                     +------- F: Failure flag:   1 = Access failed, 0 = Correct       //
//                                                                                      //
//                            DTF-Values    |     Type                                  //
//                       -------------------+----------------                           //
//                        0b11000000 (0xC0) | Write Response                            //
//                        0b11000001 (0xC1) | Read  Response                            //
//                        0b11000010 (0xC2) | Read  Request                             //
//                        0b11000011 (0xC3) | Write Request                             //
//                        0b11000100 (0xC4) | Write Failure                             //
//                        0b11000101 (0xC5) | Read  Failure                             //
//                                                                                      //
//    NET: Net address of the destination node.                                        //
//                                                                                      //
//    DEV: Device address of the destination node.                                     //
//                                                                                      //
//    LOC: Unsigned 24 bit area number of requested service.                           //
//                                                                                      //
//    CNT: Dependend on the type of datagramm: Number of provided (write request),     //
//          written (write response), requested (read request) or read (read response)  //
//          data bytes. Maximum allowed value here is 64 (0x40).                        //
//                                                                                      //
//    DAT: Payload. Only datagrams of type 'write request' and 'read response' have    //
//          payload bytes in section DAT. In Write responses, read requests and failed  //
//          datagrams have no DAT bytes, section DAT is skipped.                        //
//                                                                                      //
//    CRC  8-Bit Dow-CRC across all PDU bytes, except CRC.                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////| RG |//
static bool
SerGetRequest(Cy4Message* message)
{
    static UInt08  column                           = 0;
    static char    line[CFG_SERIAL_LINE_CAPACITY+1] = "";
    static UInt16  timeout                          = 0;

    message->valid = false;
    // (1) Textzeile empfangen
    bool lineComplete = false;
    char c;
    while (BOARD_SerReceiveByte((UInt08*)&c))
    {   timeout = BOARD_SetTimeout(CFG_SERIAL_INTERBYTE_TIMEOUT+1);
        // (1) Tabulatoren und Vorschbe in Leerzeichen umwandeln
        if (c == '\t' || c == '\v' || c == '\f') c = ' ';
        // (2) Leerzeichen zu Beginn einer Zeile berlesen
        if (column == 0 && c == ' ') continue;
        // (3) Bei zwei aufeinander folgenden Leerzeichen, das zweite ignorieren
        if (c == ' ' && line[column-1] == ' ') continue;
        // (4) Zeilenvorschub und Wagenrcklauf in '\0' umwandeln
        if (c == '\n' || c == '\r') c = '\0';
        // (5) Leere Zeilen ignorieren
        if (column == 0 && c == '\0') continue;
        // (6) Zeichen im Zeilenpuffer ablegen
        if (column < sizeof(line))
        {   line[column] = c;
            column++;
        }
        else
        {   line[0] = '\0'; // Zeile bergelaufen -> Zeile leer
        }
        // (7) Zeilenende erreicht
        if (c == '\0')
        {   if (line[column-1] == ' ') line[column-1] = '\0'; // Ggf. letztes ' ' entfernen
            column = 0;
            lineComplete = true;
            break;
        }
    }
    if (BOARD_Timeout(timeout)) column = 0;
    if (!lineComplete) return false;
	// (2) Prfen, ob das Asterisk-Anfrage gegeben wurde
    if (line[0] == '*' && line[1] == '\0')
    {   // Gerteinformation ausgeben
        CY4MESSAGE_SetupRead(message, selfAdr, 0xFFFFF0, 7);
        return true;
    }
    // (3) Anfrage-Nachricht aus Textzeile extrahieren
    const char* text = line;
    UInt08      crc  = 0xFF;
    UInt08      buffer[7];
    for (UInt08 i=0; i<7; i++)
    {   if (!ParseUInt08Hex(&text, &buffer[i])) return false;
        UpdateCRC(buffer[i], &crc);
    }
    if (buffer[0] < 0xC2 || buffer[0] > 0xC3) return false;    // Requests only
    bool payload = buffer[0] & 0x01;
    if (payload) message->access = CY4MESSAGE_WRITE;
    else          message->access = CY4MESSAGE_READ;
    message->address.net = buffer[1];
    message->address.dev = buffer[2];
    message->location    = ((UInt32)buffer[3]<<16) + ((UInt16)buffer[4]<<8) + buffer[5];
    message->count       = buffer[6];
    if (message->count > PAYLOAD_CAPACITY) return false;
    UInt08  data;
    if (payload)
    {   UInt08 i;
        for (i=0; i<message->count; i++)
        {   if (!ParseUInt08Hex(&text, &data)) return false;
            message->data[i] = data;
            UpdateCRC(data, &crc);
        }
        i += 7;
    }
    if (!ParseUInt08Hex(&text, &data)) return false;
    if (*text != '\0' || crc != data) return false;
    message->valid = message->address.net == selfAdr.net &&
                     message->address.dev == selfAdr.dev;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SerPutResponse(const Cy4Message* message)
{
    // (1) Antwort als Textzeile exportieren
    char line[CFG_SERIAL_LINE_CAPACITY+1];
    char* text = line;
    bool payload   = message->access == CY4MESSAGE_READ;
    bool direction = false; // Response
    bool failure   = !message->valid;
    UInt08 buffer[7];
    buffer[0] = 0xC0 + (failure<<2) + (direction<<1) + (payload<<0);
    buffer[1] = message->address.net;
    buffer[2] = message->address.dev;
    buffer[3] = (UInt08)(message->location >> 16);
    buffer[4] = (UInt08)(message->location >>  8);
    buffer[5] = (UInt08)(message->location >>  0);
    buffer[6] = message->count;
    UInt08 crc = 0xFF;
    for (UInt08 i=0; i<7; i++)
    {   PrintUInt08Hex(&text, buffer[i]);
        UpdateCRC(buffer[i], &crc);
    }
    if (!failure && payload)
    {   for (UInt08 i=0; i<buffer[6]; i++)
        {   PrintUInt08Hex(&text, message->data[i]);
            UpdateCRC(message->data[i], &crc);
        }
    }
    PrintUInt08Hex(&text, crc);
    // (2) Textzeile bertragen
    text = line;
    while (*text != '\0')
    {   BOARD_SerTransmitByte(*text);
        text++;
    }
    char newLine[] = CFG_SERIAL_NEWLINE;
    text = newLine;
    while (*text != '\0')
    {   BOARD_SerTransmitByte(*text);
        text++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProcessMsg(Cy4Message* message)
{
    static UInt16  timer = 0;

    if (BOARD_Timeout(timer))
    {   // Wenn lngere Zeit kein Datagramm eintrifft, Programmierung abbrechen
        progInProg = false;
    }
    if (!message->valid) return;
    timer = BOARD_SetTimeout(CFG_DEAD_PEER_TIMEOUT);
    message->valid = false;
    UInt08 count    = message->count;
    UInt32 location = message->location;
    if (location < PGMLDR_START)
    {   // Programming message
        if (message->access == CY4MESSAGE_WRITE)
        {   if (!progInProg)
            {   progInProg = true;
                ramPageNo = 0;
                ReadFlashPage(0);
                boot_page_erase_safe(0);
            }
            for (UInt08 i=0; i<count; i++)
            {   UInt16 accPageNo = (UInt16)(location >> CFG_PAGE_SIZE);
                if (accPageNo != ramPageNo)
                {   if (ramPageNo > 0)
                    {   if (!WriteFlashPage(ramPageNo)) return;  // Schreibfehler
                    }
                    if (accPageNo > 0) ReadFlashPage(accPageNo);
                    ramPageNo = accPageNo;
                }
                UInt08 index = ((UInt08)location) & ((1<<CFG_PAGE_SIZE)-1);
                if (ramPageNo == 0) zero[index] = message->data[i];
                else                 page[index] = message->data[i];
                location++;
            }
        }
    }
    else
    {   if (location < 0xFFFFF0) return;
        // System message
        UInt08 locLoByte = (UInt08)location;
        if (message->access == CY4MESSAGE_READ)
        {   // Get type plate
            UInt08 i = 0;
            UInt08 k = (locLoByte-0xF0);
            do
	        {   static const UInt08 typePlate[][7] PROGMEM = { { 'C','y','4','C','o','d','e' },
                                                               CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                            DEV_APPL_VERS,
                                                                            DEV_BOARD_TYPE)
                                                             };
                if (k >= CARDINALITY(typePlate[1])) return;
                message->data[i] = pgm_read_byte(&typePlate[1][k]);
                k++;
                i++;
            } while (i<count);
        }
        else // CY4MESSAGE_WRITE
        {   if (count == 1                               &&
                (locLoByte == 0xF8 || locLoByte == 0xFA) &&
                message->data[0] == '^')
            {   // Enter application / restart programm loader
                if (locLoByte == 0xF8 && progInProg)
                {   WriteFlashPage(ramPageNo);
                    WriteFlashPage(0);
                }
                if (locLoByte == 0xFA) BOARD_Restart();
                else                   BOARD_EnterAppl();
            }
            else if (count == 2                              &&
                     locLoByte == 0xFC                       &&
                     (message->data[0] != 0 && message->data[1] != 0))
            {   // Change device address
                Cy4Address newAdress;
                newAdress.net = message->data[0];
                newAdress.dev = message->data[1];
                SaveSelfAddress(newAdress);
                BOARD_Restart();
            }
            else
            {   return;
            }
        }
    }
    message->valid = true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Hauptschleife  ///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
main(void)
{
    Cy4Message  message;

    LoadSelfAddress();
    BOARD_Startup();
    while (BOARD_KeepRunning())
    {   BOARD_Proceed();
        BusGetRequest(&message);
        ProcessMsg(&message);
        BusPutResponse(&message);
        if (SerGetRequest(&message))
        {   ProcessMsg(&message);
            SerPutResponse(&message);
        }
    }
    BOARD_Shutdown();
}

//////////////////////////////////////////////////////////////////////////////////////////
